﻿
using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;
using System.Reflection;
using System.Runtime.CompilerServices;

var b = new AllBenchmarks();
b.GetMethodInfoFromName();
b.CallViaMethodInfo();

BenchmarkRunner.Run<AllBenchmarks>();

[MemoryDiagnoser]
public class AllBenchmarks {

	public static int IntField;

	[MethodImpl(MethodImplOptions.NoInlining)]
	public static string Concat(string a, string b) => a + b;

	[MethodImpl(MethodImplOptions.NoInlining)]
	public static int Sum(int a, int b) => a + b;

	[MethodImpl(MethodImplOptions.NoInlining)]
	public int SumInstance(int a, int b) => a + b;

	private Func<string, string, string> concatDelegate;
	private Func<int, int, int> sumDelegate;
	private Func<int, int, int> sumInstanceDelegate;
	private MethodInfo sumMethodInfo;
	private Func<int, int, int> sumDelegateFromMethodInfo;

	private FieldInfo fieldInfo;

	public AllBenchmarks() {
		concatDelegate = new Func<string, string, string>(Concat);
		sumDelegate = new Func<int, int, int>(Sum);
		sumInstanceDelegate = new Func<int, int, int>(this.SumInstance);
		sumMethodInfo = typeof(AllBenchmarks).GetMethod(nameof(Sum), BindingFlags.Static | BindingFlags.Public)!;
		sumDelegateFromMethodInfo = sumMethodInfo.CreateDelegate<Func<int, int, int>>();

		fieldInfo = typeof(AllBenchmarks).GetField(nameof(IntField), BindingFlags.Static | BindingFlags.Public)!;
	}

	[Benchmark]
	public string Concat_CallStaticMethod() {
		return Concat("Hello", "World");
	}

	[Benchmark]
	public Func<string, string, string> Concat_CreateDelegateStaticMethod() {
		return new Func<string, string, string>(Concat);
	}

	[Benchmark]
	public string Concat_CallStaticMethodDelegate() {
		return concatDelegate("Hello", "World");
	}

	[Benchmark]
	public int Sum_CallStaticMethod() {
		return Sum(10, 20);
	}

	[Benchmark]
	public Func<int, int, int> Sum_CreateDelegateStaticMethod() {
		return new Func<int, int, int>(Sum);
	}

	[Benchmark]
	public int Sum_CallStaticMethodDelegate() {
		return sumDelegate(10, 20);
	}

	[Benchmark]
	public int CallInstanceMethod() {
		return this.SumInstance(10, 20);
	}

	[Benchmark]
	public Func<int, int, int> CreateDelegateInstanceMethod() {
		return new Func<int, int, int>(this.SumInstance);
	}

	[Benchmark]
	public int CallInstanceMethodDelegate() {
		return sumInstanceDelegate(10, 20);
	}

	[Benchmark]
	public MethodInfo GetCachedDelegateMethodInfo() {
		return sumDelegate.Method;
	}

	[Benchmark]
	public MethodInfo GetNewDelegateMethodInfo() {
		var d = new Func<int, int, int>(Sum);
		return d.Method;
	}

	[Benchmark]
	public MethodInfo? GetMethodInfoFromName() {
		var mi = typeof(AllBenchmarks).GetMethod(nameof(Sum), BindingFlags.Static | BindingFlags.Public);
		return mi;
	}

	[Benchmark]
	public int CallViaMethodInfo() {
		object? result = sumMethodInfo.Invoke(null, new object[] { 10, 20 });
		return (int) result!;
	}

	[Benchmark]
	public Func<int, int, int> CreateDelegateFromMethodInfo() {
		return sumMethodInfo.CreateDelegate<Func<int, int, int>>();
	}

	[Benchmark]
	public int CallDelegateFromMethodInfo() {
		return sumDelegateFromMethodInfo(10, 20);
	}

	[Benchmark]
	public void SetStaticField() {
		IntField = 42;
	}

	[Benchmark]
	public void SetStaticFieldViaSetValue() {
		fieldInfo.SetValue(null, 42);
	}
}
