﻿using System.Text.Json.Serialization.Metadata;
using System.Text.Json;
using System.Reflection;

namespace SerializationWithTextJson {

    record Person(string FirstName, string LastName, int Age);

    class Child {
        private string _name;
        private int _age;

        public Child(string name, int age) {
            _name = name;
            _age = age;
        }

        public override string ToString() => $"Child({_name}, {_age})";
    }

    internal class Program {
        static void Main(string[] args) {
            var p = new Person(FirstName: "John", LastName: "Doe", Age: 22);

            Console.WriteLine("Person instance in JSON:");
            Console.WriteLine(JsonSerializer.Serialize(p));
            Console.WriteLine();

            Console.WriteLine("Person instance in JSON (WriteIndented = true):");
            var basicOptions = new JsonSerializerOptions { WriteIndented = true };
            Console.WriteLine(JsonSerializer.Serialize(p, basicOptions));
            Console.WriteLine();

            var c = new Child(name: "Jack", age: 12);

            Console.WriteLine("Child instance in JSON:");
            Console.WriteLine(JsonSerializer.Serialize(c, basicOptions));
            Console.WriteLine();

            var optionsWithPrivateFields = new JsonSerializerOptions {
                WriteIndented = true,
                TypeInfoResolver = new DefaultJsonTypeInfoResolver {
                    Modifiers = { AddPrivateFieldsModifier }
                }
            };

            Console.WriteLine("Child instance in JSON (custom TypeInfoResolver.Modifiers to include private fields):");
            Console.WriteLine(JsonSerializer.Serialize(c, optionsWithPrivateFields));
            Console.WriteLine();

            Console.WriteLine(JsonSerializer.Serialize(p, optionsWithPrivateFields));

            Console.WriteLine();

            string json = JsonSerializer.Serialize(p);
            Person? p2 = JsonSerializer.Deserialize<Person>(json);

            if (p2 is not null) {
                Console.WriteLine(p == p2);
                Console.WriteLine(ReferenceEquals(p, p2));
            }
        }

        static void AddPrivateFieldsModifier(JsonTypeInfo jsonTypeInfo) {
            if (jsonTypeInfo.Kind != JsonTypeInfoKind.Object)
                return;

            var fields = jsonTypeInfo.Type.GetFields(BindingFlags.Instance | BindingFlags.NonPublic);
            foreach (FieldInfo field in fields) {
                JsonPropertyInfo jsonPropertyInfo = jsonTypeInfo.CreateJsonPropertyInfo(field.FieldType, field.Name);
                jsonPropertyInfo.Get = field.GetValue;
                jsonPropertyInfo.Set = field.SetValue;

                jsonTypeInfo.Properties.Add(jsonPropertyInfo);
            }
        }

    }
}